#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shlwapi.h>
#include <yuna_if.h>
#include <cstdio>
#include <cstring>
#include "rena_api.h"

using namespace std;

#define RENA_TYPE_CLIENT  (0)
#define RENA_TYPE_SERVER  (1)

//Global data
static bool g_fail = false;
static bool g_boot = false;
static HMODULE g_module = NULL;
static HANDLE g_mail_client = INVALID_HANDLE_VALUE;
static HANDLE g_mail_server = INVALID_HANDLE_VALUE;
static HANDLE g_mail_cl_res = INVALID_HANDLE_VALUE;
static HANDLE g_mail_sv_res = INVALID_HANDLE_VALUE;

//Entry point
int APIENTRY DllMain(HANDLE module, DWORD, LPVOID) {
	g_module = (HMODULE)module;
	return TRUE;
}

//Boot YUNA
static int rena_boot_yuna(void) {
	//NOP if already booted or failed
	if (g_boot || g_fail) {
		return RENA_OK;
	}

	//Make YUNA module path
	char path[MAX_PATH * 2];
	::GetModuleFileName(g_module, path, sizeof(path));
	strcpy(::PathFindFileName(path), "\\yuna.exe");

	//Boot YUNA
	STARTUPINFO si = { sizeof(STARTUPINFO) };
	PROCESS_INFORMATION pi;
	BOOL result = ::CreateProcess(path, NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi);

	//Check result
	g_boot = (result == TRUE);
	g_fail = (result == FALSE);
	return g_fail ? RENA_NG : RENA_OK;
}

//Open mailslot
static int rena_open_mailslot(const char *p_prefix, HANDLE *p_target) {
	HANDLE result = INVALID_HANDLE_VALUE;
	unsigned long timeout = ::GetTickCount() + (10 * 1000);

	//Make sender name
	char name[256];
	sprintf(name, "\\\\.\\mailslot\\com.lmt.windows.yuna.%s", p_prefix);

	//Creation loop
	while ((result == INVALID_HANDLE_VALUE) && (::GetTickCount() < timeout)) {
		result = ::CreateFile(name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (result == INVALID_HANDLE_VALUE) {
			::Sleep(10);
		}
	}
	if (result == INVALID_HANDLE_VALUE) {
		g_fail = true;
	}

	//OK?
	*p_target = result;
	return g_fail ? RENA_NG : RENA_OK;
}

//Create mailslot for response
static int rena_create_response_mailslot(const char *p_prefix, unsigned long timeout, HANDLE sender, HANDLE *p_target) {
	//Fail if sender is nothing
	if (sender == INVALID_HANDLE_VALUE) {
		g_fail = true;
		return RENA_NG;
	}

	//Make mailslot name
	char name[256];
	SYSTEMTIME st;
	::GetLocalTime(&st);
	sprintf(name, "\\\\.\\mailslot\\com.lmt.windows.yuna.%s.response.%04u%02u%02u%02u%02u%02u%03u",
		p_prefix, st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds);

	//Try create mailslot
	HANDLE handle = ::CreateMailslot(name, YUNA_MESSAGE_SIZE, timeout, NULL);
	if (handle == INVALID_HANDLE_VALUE) {
		g_fail = true;
		return RENA_NG;
	}

	//Notify mailslot name
	YUNA_MSG msg;
	YUNA_MAIL *p_mail = (YUNA_MAIL*)msg.info;
	unsigned long wrote;

	msg.kind = YUNA_MSG_KIND_NOTIFY_MAIL;
	strcpy(p_mail->name, name);
	::WriteFile(sender, &msg, sizeof(msg), &wrote, NULL);

	*p_target = handle;
	return RENA_OK;
}

//Connect to YUNA
static int rena_connect_yuna(const char *p_prefix, unsigned long timeout, HANDLE *p_sender, HANDLE *p_receiver) {
	if (g_fail) {
		return RENA_NG;
	}

	//Setup sender
	if (*p_sender == INVALID_HANDLE_VALUE) {
		::rena_open_mailslot(p_prefix, p_sender);
	}

	//Setup receiver
	if (*p_receiver == INVALID_HANDLE_VALUE) {
		::rena_create_response_mailslot(p_prefix, timeout, *p_sender, p_receiver);
	}

	//Success?
	return g_fail ? RENA_NG : RENA_OK;
}

//Initialize base
static int rena_init(int type) {
	//Decide data
	const char *p_prefix;
	unsigned long timeout;
	HANDLE *p_sender;
	HANDLE *p_receiver;
	if (type == RENA_TYPE_CLIENT) {
		p_prefix = "client";
		timeout = MAILSLOT_WAIT_FOREVER;
		p_sender = &g_mail_client;
		p_receiver = &g_mail_cl_res;
	}
	else if (type == RENA_TYPE_SERVER) {
		p_prefix = "server";
		timeout = 1000;
		p_sender = &g_mail_server;
		p_receiver = &g_mail_sv_res;
	}
	else {
		return RENA_NG;
	}

	//Execute initialize
	if (::rena_boot_yuna() == RENA_NG) {
		return RENA_NG;
	}
	if (::rena_connect_yuna(p_prefix, timeout, p_sender, p_receiver) == RENA_NG) {
		return RENA_NG;
	}

	//Good
	return RENA_OK;
}

//Send log
int __stdcall rena_send(int level, const char *p_tag, const char *p_msg) {
	//Initialize
	if (::rena_init(RENA_TYPE_CLIENT) == RENA_NG) {
		return RENA_NG;
	}

	//Make packet
	YUNA_MSG msg;
	YUNA_LOG *p_log = (YUNA_LOG*)msg.info;
	unsigned int tag_len = strlen(p_tag);
	unsigned int msg_len = strlen(p_msg);

	//Adjust length
	if (tag_len > RENA_LENGTH_TAG) {
		tag_len = RENA_LENGTH_TAG;
	}
	if (msg_len > RENA_LENGTH_MSG) {
		msg_len = RENA_LENGTH_MSG;
	}

	msg.kind = YUNA_MSG_KIND_SEND_LOG_REQ;
	p_log->tick = ::GetTickCount();
	p_log->level = level;
	memcpy(p_log->tag, p_tag, tag_len);
	p_log->tag[tag_len] = '\0';
	memcpy(p_log->msg, p_msg, msg_len);
	p_log->msg[msg_len] = '\0';

	//Send SEND_LOG_REQ
	unsigned long wrote;
	::WriteFile(g_mail_client, &msg, sizeof(msg), &wrote, NULL);

	//Receive SEND_LOG_RES
	YUNA_RES res;
	unsigned long read;
	BOOL result = ::ReadFile(g_mail_cl_res, &res, sizeof(res), &read, NULL);

	return (result && res.res == YUNA_RES_OK) ? RENA_OK : RENA_NG;
}

//Receive log
int __stdcall rena_recv(unsigned long *p_tick, int *p_level, char *p_tag, char *p_msg) {
	//Initialize
	if (::rena_init(RENA_TYPE_SERVER) == RENA_NG) {
		return RENA_NG;
	}

	//Make packet
	YUNA_MSG msg;
	msg.kind = YUNA_MSG_KIND_RECV_LOG_REQ;

	//Send RECV_LOG_REQ
	unsigned long wrote;
	::WriteFile(g_mail_server, &msg, sizeof(msg), &wrote, NULL);

	//Receive RECV_LOG_RES
	YUNA_RES res;
	YUNA_LOG *p_log = (YUNA_LOG*)res.info;
	unsigned long read;
	BOOL result = ::ReadFile(g_mail_sv_res, &res, sizeof(res), &read, NULL);
	if ((!result) || (res.kind != YUNA_MSG_KIND_RECV_LOG_RES) || (res.res != YUNA_RES_OK)) {
		return RENA_NG;
	}

	//Copy to user buffer
	*p_tick = p_log->tick;
	*p_level = p_log->level;
	strcpy(p_tag, p_log->tag);
	strcpy(p_msg, p_log->msg);
	return RENA_OK;
}

//Set pause flag
int __stdcall rena_set_pause(int pause) {
	//Initialize
	if (::rena_init(RENA_TYPE_SERVER) == RENA_NG) {
		return RENA_NG;
	}

	YUNA_MSG msg;
	YUNA_PAUSE *p_pause = (YUNA_PAUSE*)msg.info;
	YUNA_RES res;
	unsigned long wrote;
	unsigned long read;

	//Send SET_PAUSE_REQ
	msg.kind = YUNA_MSG_KIND_SET_PAUSE_REQ;
	p_pause->pause = pause;
	::WriteFile(g_mail_server, &msg, sizeof(msg), &wrote, NULL);

	//Receive SET_PAUSE_RES
	BOOL result = ::ReadFile(g_mail_sv_res, &res, sizeof(res), &read, NULL);
	return (result && res.res == YUNA_RES_OK) ? RENA_OK : RENA_NG;
}